import json
import sys
from dataclasses import asdict
from pathlib import Path
from typing import List, Optional

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated  # type: ignore

import stlog
import tomli
import typer

from jinja_tree.app.config import (
    Config,
    make_default_action_plugin_config,
    make_default_context_plugin_config,
)
from jinja_tree.infra.utils import get_config_file_path

ConfigFileType = Annotated[
    Optional[str],
    typer.Option(
        help="config file path (default: first '.jinja-tree.toml' file found up from current working dir), can also be see with JINJA_TREE_CONFIG_FILE env var",
        envvar="JINJA_TREE_CONFIG_FILE",
    ),
]
LogLevelType = Annotated[
    Optional[str], typer.Option(help="log level (DEBUG, INFO, WARNING or ERROR)")
]

ExtraSearchPathsType = Annotated[
    Optional[List[Path]], typer.Option(help="Search path to jinja")
]

AddProcessedFileDirToSearchPathType = Annotated[
    Optional[bool],
    typer.Option(help="add directory of current processed file to jinja search path"),
]

AddRootDirToSearchPathType = Annotated[
    Optional[bool], typer.Option(help="add root directory to jinja search path")
]

AddCwdDirToSearchPathType = Annotated[
    Optional[bool],
    typer.Option(help="add current working directory (CWD) to jinja search path"),
]

ExtensionType = Annotated[
    Optional[List[str]], typer.Option(help="jinja extension to load")
]

ContextPluginType = Annotated[
    Optional[str], typer.Option(help="context plugin (full python class path)")
]

StrictUndefinedType = Annotated[
    Optional[bool],
    typer.Option(help="if set, raise an error if a variable does not exist in context"),
]

FileActionPluginType = Annotated[
    Optional[str], typer.Option(help="action plugin (full python class path)")
]

BlankRunType = Annotated[
    bool,
    typer.Option(
        help="if set, execute a blank run (without modifying or deleting anything)"
    ),
]

DisableEmbeddedExtensions = Annotated[
    Optional[bool], typer.Option(help="disable embedded jinja extensions")
]


RootDirType = Annotated[Path, typer.Argument(help="root directory")]


def get_config(
    config_file_path: ConfigFileType = None,
    extra_search_path: ExtraSearchPathsType = None,
    add_cwd_to_search_path: AddCwdDirToSearchPathType = None,
    add_root_dir_to_search_path: AddRootDirToSearchPathType = None,
    strict_undefined: StrictUndefinedType = None,
    jinja_extension: ExtensionType = None,
    disable_embedded_jinja_extensions: DisableEmbeddedExtensions = None,
    root_dir: Optional[RootDirType] = None,
    context_plugin: ContextPluginType = None,
    action_plugin: FileActionPluginType = None,
) -> Config:
    if not config_file_path:
        config_file_path = get_config_file_path()
    general = {}
    context_plugin_config = make_default_context_plugin_config()
    action_plugin_config = make_default_action_plugin_config()
    if config_file_path:
        with open(config_file_path, "rb") as f:
            data = tomli.load(f)
        general = data.get("general", {})
        context_plugin_config = {**context_plugin_config, **data.get("context", {})}
        action_plugin_config = {
            **action_plugin_config,
            **data.get("action", {}),
        }
    config = Config(
        **general,
        context_plugin_config=context_plugin_config,
        action_plugin_config=action_plugin_config,
    )
    if extra_search_path:
        config.extra_search_paths = [str(x) for x in extra_search_path]
    if add_cwd_to_search_path is not None:
        config.add_cwd_to_search_path = add_cwd_to_search_path
    if add_root_dir_to_search_path is not None:
        config.add_root_dir_to_search_path = add_root_dir_to_search_path
    if strict_undefined is not None:
        config.strict_undefined = strict_undefined
    if jinja_extension:
        config.jinja_extensions = jinja_extension
    if disable_embedded_jinja_extensions is not None:
        config.disable_embedded_jinja_extensions = disable_embedded_jinja_extensions
    if root_dir is not None:
        config.root_dir = str(root_dir)
    if context_plugin is not None:
        config.context_plugin_config["plugin"] = context_plugin
    if action_plugin is not None:
        config.action_plugin_config["plugin"] = action_plugin
    config.__post_init__()
    return config


def config_dump(config: Config):
    print("<config dump>", file=sys.stderr)
    print(json.dumps(asdict(config), indent=4, sort_keys=True), file=sys.stderr)
    print("</config dump>", file=sys.stderr)


def setup_logger(log_level: Optional[str]):
    if log_level is None:
        log_level = "INFO"
    stlog.setup(level=log_level)



import fnmatch
import os
from importlib import import_module
from typing import List, Optional, Type

import stlog

from jinja_tree.app.action import ActionPort
from jinja_tree.app.config import (
    CONTEXT_PLUGIN_DEFAULT,
    FILE_ACTION_PLUGIN_DEFAULT,
    Config,
)
from jinja_tree.app.context import ContextPort

SYSTEM_CONFIG_PATH = "/etc/jinja-tree.toml"

logger = stlog.getLogger("jinja_tree")


def import_class_from_string(class_path: str) -> Type:
    module_name, _, class_name = class_path.rpartition(".")
    klass = getattr(import_module(module_name), class_name)
    return klass


def make_context_adapter_from_config(config: Config) -> ContextPort:
    class_path = config.context_plugin_config.get("plugin", CONTEXT_PLUGIN_DEFAULT)
    context_adapter_class = import_class_from_string(class_path)
    context_adapter = context_adapter_class(config=config)
    if not isinstance(context_adapter, ContextPort):
        raise Exception(
            f"the class pointed by {class_path} does not implement ContextPort interface"
        )
    return context_adapter


def make_file_action_adapter_from_config(config: Config) -> ActionPort:
    class_path = config.action_plugin_config.get("plugin", FILE_ACTION_PLUGIN_DEFAULT)
    context_adapter_class = import_class_from_string(class_path)
    file_action_adapter = context_adapter_class(config=config)
    if not isinstance(file_action_adapter, ActionPort):
        raise Exception(
            f"the class pointed by {class_path} does not implement FileActionPort interface"
        )
    return file_action_adapter


def get_config_file_path(
    cli_option: Optional[str] = None, cwd: Optional[str] = None
) -> Optional[str]:
    """
    Get the path to the configuration file.

    - if the CLI option is set, it is returned immediately (note: can be set by JINJA_TREE_CONFIG_FILE env var)
    - else, we will try to find it in the current working directory
        or in the parent directory (recursively).

    Args:
        cli_option: The configuration file path provided via CLI
            (if set, the function returns this value immediately)
        cwd: The current working directory (if not set, the current working directory is used)

    Returns:
        The path to the configuration file, or None if not found.

    """
    if cli_option is not None:
        logger.debug(f"config file path read from CLI: {cli_option}")
        return cli_option
    if cwd is None:
        cwd = os.getcwd()
    else:
        cwd = os.path.abspath(cwd)
    path = os.path.join(cwd, ".jinja-tree.toml")
    if os.path.isfile(path):
        logger.debug(f"config file path found in: {path}")
        return os.path.abspath(path)
    parent_path = os.path.dirname(cwd)
    if parent_path == cwd:
        # we are done
        # let's try the system config path?
        if os.path.isfile(SYSTEM_CONFIG_PATH):
            return SYSTEM_CONFIG_PATH
        logger.debug("no config file found")
        return None
    return get_config_file_path(cwd=parent_path)


def is_fnmatch_ignored(key: str, ignores: List[str]) -> bool:
    """
    Check if the given key matches any of the patterns in the ignores list using fnmatch.

    Args:
        key: The key to check.
        ignores: The list of patterns to ignore.

    Returns:
        True if the key matches any of the patterns, False otherwise.
    """
    return any(fnmatch.fnmatch(key, x) for x in ignores)



import stlog
import typer
import typer.core

from jinja_tree.app.action import ActionService
from jinja_tree.app.context import ContextService
from jinja_tree.app.jinja import JinjaService
from jinja_tree.app.jinja_tree import JinjaTreeService
from jinja_tree.infra.controllers.cli_common import (
    AddCwdDirToSearchPathType,
    AddRootDirToSearchPathType,
    BlankRunType,
    ConfigFileType,
    ContextPluginType,
    DisableEmbeddedExtensions,
    ExtensionType,
    ExtraSearchPathsType,
    FileActionPluginType,
    LogLevelType,
    RootDirType,
    StrictUndefinedType,
    config_dump,
    get_config,
    setup_logger,
)
from jinja_tree.infra.utils import (
    make_context_adapter_from_config,
    make_file_action_adapter_from_config,
)

# disable rich usage in typer
typer.core.rich = None  # type: ignore
app = typer.Typer(add_completion=False)


@app.command()
def tree(
    root_dir: RootDirType,
    config_file: ConfigFileType = None,
    log_level: LogLevelType = "INFO",
    extra_search_path: ExtraSearchPathsType = None,
    add_cwd_to_search_path: AddCwdDirToSearchPathType = None,
    add_root_dir_to_search_path: AddRootDirToSearchPathType = None,
    jinja_extension: ExtensionType = None,
    context_plugin: ContextPluginType = None,
    action_plugin: FileActionPluginType = None,
    strict_undefined: StrictUndefinedType = None,
    blank_run: BlankRunType = False,
    disable_embedded_jinja_extensions: DisableEmbeddedExtensions = None,
):
    """
    Process a directory tree with the Jinja / Jinja2 templating system.

    """
    if log_level is not None:
        stlog.setup(level=log_level)
    config = get_config(
        config_file_path=config_file,
        extra_search_path=extra_search_path,
        add_cwd_to_search_path=add_cwd_to_search_path,
        add_root_dir_to_search_path=add_root_dir_to_search_path,
        jinja_extension=jinja_extension,
        strict_undefined=strict_undefined,
        context_plugin=context_plugin,
        action_plugin=action_plugin,
        disable_embedded_jinja_extensions=disable_embedded_jinja_extensions,
        root_dir=root_dir,
    )
    if log_level == "DEBUG":
        config_dump(config)
    setup_logger(log_level)
    context_adapter = make_context_adapter_from_config(config)
    file_action_adapter = make_file_action_adapter_from_config(config)
    context_service = ContextService(config=config, adapter=context_adapter)
    file_action_service = ActionService(config=config, adapter=file_action_adapter)
    jinja_service = JinjaService(config=config, context_service=context_service)
    jinja_tree_service = JinjaTreeService(
        config=config,
        action_service=file_action_service,
        jinja_service=jinja_service,
        blank_run=blank_run,
    )
    jinja_tree_service.process()


if __name__ == "__main__":
    app()



import os
import shutil

import stlog

from jinja_tree.app.action import (
    ActionService,
    IgnoreDirectoryAction,
    IgnoreFileAction,
    ProcessFileAction,
    RenameFileAction,
)
from jinja_tree.app.config import Config
from jinja_tree.app.jinja import JinjaService

logger = stlog.getLogger("jinja-tree")


def get_source_content(source_absolute_path: str) -> str:
    with open(source_absolute_path) as f:
        return f.read()


def get_target_content(target_absolute_path: str) -> str:
    with open(target_absolute_path) as f:
        return f.read()


def set_target_content(
    target_absolute_path: str, content: str, blank_run: bool = False
):
    if (
        os.path.isfile(target_absolute_path)
        and get_target_content(target_absolute_path) == content
    ):
        logger.info("No change to apply to the target file", path=target_absolute_path)
        return
    if blank_run:
        print(f"[BLANK RUN] Fake setting new content for {target_absolute_path}")
        print("[BLANK RUN]<new_content>")
        print(content)
        print("[BLANK RUN]</new_content>")
        return
    logger.info("New content written to target file", path=target_absolute_path)
    with open(target_absolute_path, "w") as f:
        f.write(content)


def delete_source_file(filepath: str, blank_run: bool = False):
    if blank_run:
        print(f"[BLANK RUN] Fake deleting {filepath}...")
    else:
        logger.info("Deleting a file", path=filepath)
        os.unlink(filepath)


def chdir(dirpath: str):
    logger.debug("Changing working directory", new_path=dirpath)
    os.chdir(dirpath)


class JinjaTreeService:
    """
    A service class for processing tree of files using Jinja templates.

    Args:
        config: The configuration object.
        file_action_service: The FileAction service.
        jinja_service: The JinjaService object to render single strings.
        blank_run: Flag indicating if it is a blank run.
    """

    def __init__(
        self,
        config: Config,
        action_service: ActionService,
        jinja_service: JinjaService,
        blank_run: bool = False,
    ):
        self.config = config
        self.action_service = action_service
        self.jinja_service = jinja_service
        self.blank_run = blank_run

    def do_process_action(self, action: ProcessFileAction):
        if self.config.change_cwd:
            chdir(action.dirpath)
        content = get_source_content(action.source_absolute_path)
        logger.info(
            "Processing a file",
            path=action.source_absolute_path,
            target=action.target_absolute_path,
        )
        result = self.jinja_service.render_string(content, action.source_absolute_path)
        set_target_content(
            action.target_absolute_path, result, blank_run=self.blank_run
        )
        if action.delete_original:
            delete_source_file(action.source_absolute_path, blank_run=self.blank_run)

    def do_rename_action(self, action: RenameFileAction):
        if self.blank_run:
            print(
                f"[BLANK RUN] Fake renaming {action.source_absolute_path} to {action.target_absolute_path}..."
            )
        else:
            logger.info(
                f"Renaming {action.source_absolute_path} to {action.target_absolute_path}..."
            )
            shutil.move(action.source_absolute_path, action.target_absolute_path)

    def process(self):
        for dirpath, dirnames, filenames in os.walk(self.config.root_dir, topdown=True):
            # Directory action
            dir_action = self.action_service.get_directory_action(dirpath)
            if isinstance(dir_action, IgnoreDirectoryAction):
                # modify in-place, see https://stackoverflow.com/a/19859907
                # to ignore recursively
                dirnames[:] = []
                continue

            # File actions
            for f in filenames:
                path = os.path.join(dirpath, f)
                action = self.action_service.get_file_action(path)
                if isinstance(action, IgnoreFileAction):
                    pass
                elif isinstance(action, ProcessFileAction):
                    self.do_process_action(action)
                elif isinstance(action, RenameFileAction):
                    self.do_rename_action(action)
                else:
                    raise Exception("unknown action type: %s", action)



import os
from typing import Any, Dict, List, cast

from dotenv import dotenv_values

from jinja_tree.app.config import Config
from jinja_tree.app.context import ContextPort
from jinja_tree.infra.utils import is_fnmatch_ignored


class EnvContextAdapter(ContextPort):
    def __init__(self, config: Config):
        self.env_ignores = cast(
            List[str], config.context_plugin_config.get("env_ignores")
        )
        assert self.env_ignores is not None
        self.plugin_configuration_ignores = cast(
            List[str], config.context_plugin_config.get("plugin_configuration_ignores")
        )
        assert self.plugin_configuration_ignores is not None
        self.dotenv_path = config.context_plugin_config.get("dotenv_path")
        assert self.dotenv_path is not None
        self.dotenv_ignores = cast(
            List[str], config.context_plugin_config.get("dotenv_ignores")
        )
        assert self.dotenv_ignores is not None
        self.config = config

    def get_plugin_configuration_context(self) -> Dict[str, Any]:
        if self.plugin_configuration_ignores == ["*"]:
            return {}
        return {
            x: y
            for x, y in self.config.context_plugin_config.items()
            if not is_fnmatch_ignored(x, self.plugin_configuration_ignores)
        }

    def get_dotenv_context(self) -> Dict[str, Any]:
        if not self.dotenv_path:
            return {}
        if not os.path.isfile(self.dotenv_path):
            return {}
        if self.dotenv_ignores == ["*"]:
            return {}
        return {
            x: y
            for x, y in dotenv_values(self.dotenv_path).items()
            if not is_fnmatch_ignored(x, self.dotenv_ignores)
        }

    def get_env_context(self) -> Dict[str, Any]:
        if self.env_ignores == ["*"]:
            return {}
        return {
            x: y
            for x, y in os.environ.items()
            if not is_fnmatch_ignored(x, self.env_ignores)
        }

    def get_context(self) -> Dict[str, Any]:
        plugin_context = self.get_plugin_configuration_context()
        env_context = self.get_env_context()
        dotenv_context = self.get_dotenv_context()
        return {**plugin_context, **env_context, **dotenv_context}



import os
from dataclasses import dataclass, field
from typing import Any, Dict, List

CONTEXT_PLUGIN_DEFAULT = "jinja_tree.infra.adapters.context.EnvContextAdapter"
FILE_ACTION_PLUGIN_DEFAULT = (
    "jinja_tree.infra.adapters.action.ExtensionsFileActionAdapter"
)
FILENAME_IGNORES_DEFAULT = [".*"]
DIRNAME_IGNORES_DEFAULT = [
    "venv",
    "site-packages",
    "__pypackages__",
    "node_modules",
    "__pycache__",
    ".*",
]
DOTENV_PATH_DEFAULT = ".env"
FILE_ACTION_PLUGIN_DEFAULT_EXTENSIONS = [".template"]
REPLACE_DEFAULT = True
DELETE_ORIGINAL_DEFAULT = False

EMBEDDED_EXTENSIONS = [
    "jinja_tree.app.embedded_extensions.from_json.FromJsonExtension",
    "jinja_tree.app.embedded_extensions.shell.ShellExtension",
    "jinja_tree.app.embedded_extensions.fnmatch.FnMatchExtension",
    "jinja_tree.app.embedded_extensions.double_quotes.DoubleQuotesExtension",
]

JINJA_TREE_URL = "https://github.com/fabien-marty/jinja-tree"


def make_default_context_plugin_config() -> Dict[str, Any]:
    tmp = {
        "plugin": CONTEXT_PLUGIN_DEFAULT,
        "generated_comment_line1": "*** GENERATED FILE - DO NOT EDIT ***",
        "generated_comment_line2": "This file was generated by jinja-tree (%s) from the template file: {{relative_filepath}}"
        % JINJA_TREE_URL,
        "env_ignores": [],
        "dotenv_path": DOTENV_PATH_DEFAULT,
        "dotenv_ignores": [],
    }
    tmp["plugin_configuration_ignores"] = list(tmp.keys()) + [
        "plugin_configuration_ignores"
    ]
    tmp["plugin_configuration_ignores"] = list(tmp.keys())
    return tmp


def make_default_action_plugin_config() -> Dict[str, Any]:
    return {
        "plugin": FILE_ACTION_PLUGIN_DEFAULT,
        "extensions": FILE_ACTION_PLUGIN_DEFAULT_EXTENSIONS,
        "filename_ignores": FILENAME_IGNORES_DEFAULT,
        "dirname_ignores": DIRNAME_IGNORES_DEFAULT,
        "replace": REPLACE_DEFAULT,
        "delete_original": DELETE_ORIGINAL_DEFAULT,
    }


@dataclass
class Config:
    # Global config
    extra_search_paths: List[str] = field(default_factory=list)
    add_root_dir_to_search_path: bool = True
    add_cwd_to_search_path: bool = True
    add_processed_file_dir_to_search_path: bool = False
    change_cwd: bool = True
    jinja_extensions: List[str] = field(default_factory=list)
    strict_undefined: bool = True
    root_dir: str = field(default_factory=os.getcwd)
    disable_embedded_jinja_extensions: bool = False

    # Plugin config
    context_plugin_config: Dict[str, Any] = field(
        default_factory=make_default_context_plugin_config
    )
    action_plugin_config: Dict[str, Any] = field(
        default_factory=make_default_action_plugin_config
    )

    @property
    def resolved_extensions(self) -> List[str]:
        if self.disable_embedded_jinja_extensions:
            return self.jinja_extensions
        return EMBEDDED_EXTENSIONS + self.jinja_extensions

    def __post_init__(self):
        # replaces paths by absolute paths
        self.root_dir = os.path.abspath(self.root_dir)
        self.extra_search_paths = [os.path.abspath(x) for x in self.extra_search_paths]
